// Oldham, Jeffrey D.
// 16 Apr 1995
// MIS 214
//
// Sequence Alignment Via Dynamic Programming
//
// See the README file for information.

// Header Files
#include <iostream.h>
#include <fstream.h>		// has input file stream functions
#include <String.h>		// has String class for file names
#include <GetOpt.h>		// process the optional arguments
#include <float.h>		// has minimal double number
#include "MatrixEntryPtr.OSLSet.h" // singly-linked list of table entries
	// MatrixEntryPtr.defs.h contains the definition of MatrixTable
	// Make set implementation of linked list of MatrixEntry pointers:
	//   genclass MatrixEntryPtr val OSLSet
	//   genclass MatrixEntryPtr val SLList
	//   genclass MatrixEntryPtr val Set
#include <math.h>		// has fabs()
#ifdef TIME
#include <time.h>
#endif // TIME

// Type Declarations

// SeqStruct holds all the information about a sequence.
class SeqStruct {
public:
  SeqStruct(istream & inputStream);	// read a string from the
					// input stream, storing its
					// length, and allocating the
					// intSeq memory
  ~SeqStruct();				// deallocate the allocated memory
  char * seq;				// the character sequence
					// allocated memory
  int * intSeq;				// sequence converted to
					// integral encoding
					// allocated memory
  long seqLength;			// nu of chars in the sequence
  double gapOpen;			// the penalty for starting a gap
  double gapExtend;			// the penalty for extending a
					//  gap by one character
};

// MatrixEntry holds an entry in the dynamic programming table.

struct MatrixEntry  {
  double score;				// value of match so far
  MatrixEntry * prev;			// pointer to previous match
					//  if only one match is found
  MatrixEntryPtrOSLSet prevList;		// set of best matches
					//  if more than one match desired
};

inline SeqStruct::SeqStruct(istream & inputStream)  {
  inputStream.gets(&seq);	// read the sequence
  if (!inputStream)  {  // read failed
    cerr << "Failed to read a sequence.\n";
    exit(1);
  }
  seqLength = inputStream.gcount(); // record the sequence's length
  intSeq = new int[seqLength+1];
  intSeq[seqLength] = -1;	// sentinel entry
  return;
}

inline SeqStruct::~SeqStruct()  {
  delete seq;
  delete intSeq;
  return;
}

typedef char *charPtr;


// Global Variables

bool allSequences = 0;		// holds !0 if the all best alignments
				// should be printed.  Otherwise, it
				// holds 0 and only the lowest pointer
				// alignment is printed.
bool globalAlignment = !0;	// holds !0 is the best global
				// alignment(s) should be found.
				// Otherwise, it holds 0 and the best
				// local alignment(s) should be found.
int alphabetBBits = 0;		// number of bits needed for the
				// integral representation of B's alphabet
bool endGaps = 0;		// holds !0 if end gaps should be
				// charged a gap penalty.  Only
				// applies for global alignments.  The
				// default is not to charge for end gaps.

// Constants
const char seqAGapCharacter = '_'; // sequence A's gap character in alignment
const char seqBGapCharacter = '_'; // sequence B's gap character in alignment


// Helper Functions
// Convert a two integer pair from A and B into a match location.
inline long matchLocation(int i1, int i2)
{
  return ((i1 << alphabetBBits) | i2);
}

// Return the number of bits needed to represent a number.
inline long nuBits(long nu)
{
  long nuBits;
  for (nuBits = 0; nu; nu >>= 1, nuBits++)  ;
  return nuBits;
}

// Return the maximum of two arguments.
inline double max(double arg1, double arg2)
{
  return arg1 > arg2 ? arg1 : arg2;
}

// Return the maximum of three arguments.
inline double max3(double arg1, double arg2, double arg3)
{
  return max(arg1, max(arg2, arg3));
}

// Append a copy of a linked list set "b" to another list set "a."
void copyAndAppend(MatrixEntryPtrOSLSet& a, MatrixEntryPtrOSLSet& b)
{
  Pix setPosition;		// position within set

  for (setPosition = b.first();
       setPosition /* != 0 */;
       b.next(setPosition))
    (void) a.add(b(setPosition));

  return;
}


// Main Functions

// 1. Functions to print the alignments.
//	nuRows() and nuColumns() - convert pointers to table entries to
//	 row and column positions
//	determineGapCharacters() returns the number of gap characters
//	 to print
//	printGapCharacters() adds the gap characters to the alignment
//	 strings being constructed
//	print() constructs the matched strings from table entries
//
// 2. main()
//	one giant function to read the input, fill the table, and
//	  print the matches


// Print the matched strings.

inline long nuRows(MatrixEntry * tablePtr, MatrixEntry * table,
		   long seqALength)
{ return (tablePtr - table) / seqALength; }

inline long nuColumns(MatrixEntry * tablePtr, MatrixEntry * table,
		      long seqALength)
{ return (tablePtr - table) % seqALength; }

// determineGapCharacters() returns the number of row and column gap
// characters to print.
//
// input <- tablePtr = current table entry of characters printed
//          prevTablePtr = table entry of next characters to print
//          table = beginning of array
//          seqALength = length of sequence A
//          locations for number of row and column gap characters
//           initial values are ignored
// output-> none
//	    the number of row and column gap characters to print are
//           placed in the given locations

void determineGapCharacters(MatrixEntry * tablePtr, MatrixEntry * prevTableEntry,
			    MatrixEntry * table,
			    long seqALength,
			    long &rowGapNu, long &colGapNu)
{
  if (prevTableEntry == (MatrixEntry *) 0)  {
    // Work with an entry on the leftmost column or top row.
    // Print the gap characters.
    if ((rowGapNu = nuRows(tablePtr, table, seqALength)) > 0)
      colGapNu = 0;
    else  {
      rowGapNu = 0;
      colGapNu = nuColumns(tablePtr, table, seqALength);
    }
  }
  else  {			// not left column or top row
    if ((rowGapNu = nuRows(tablePtr, prevTableEntry, seqALength) - 1) > 0)
      colGapNu = 0;
    else  {
      rowGapNu = 0;
      colGapNu = nuColumns(tablePtr, prevTableEntry, seqALength) - 1;
    }
  }
  return;
}


// printGapCharacters() prints the gap characters and the
// corresponding characters of the other sequence.
//
// input <- number of row and column gap characters to print
//          matchedAPtr = leftmost character added to the matchedA array
//	    matchedBPtr = leftmost character added to the matchedB array
//          seqAPtr = leftmost character of seqA already added to matchedAPtr
//          seqBPtr = leftmost character of seqB already added to matchedBPtr
// output-> none
//	    side effect of adding the next characters of seqA, seqB to
//            matched arrays; gap characters are also added

void printGapCharacters(int rowGapNu, int colGapNu,
			charPtr & matchedAPtr, charPtr & matchedBPtr,
			charPtr & seqAPtr, charPtr & seqBPtr)
{
  // Print the gap characters.
  for (; rowGapNu > 0; rowGapNu--)  {
    *--matchedAPtr = seqAGapCharacter;
    *--matchedBPtr = *--seqBPtr;
  }
  for (; colGapNu > 0; colGapNu--)  {
    *--matchedAPtr = *--seqAPtr;
    *--matchedBPtr = seqBGapCharacter;
  }
  return;
}


// print() adds the characters for the given table entry to the
// character arrays holding the matches.  Also, gap characters before
// these characters are added.  The alignments are printed if no more
// characters are to be added.
//
// input <- tablePtr = next "table" entry to process
//          table = dynamic table
//	    matchedAPtr = leftmost character added to the matchedA array
//	    matchedBPtr = leftmost character added to the matchedB array
//          seqAPtr = leftmost character of seqA already added to matchedAPtr
//          seqBPtr = leftmost character of seqB already added to matchedBPtr
//          seqALength = length of sequence A
// output-> none
//	    side effect of adding the next characters of seqA, seqB to
//            matched arrays; gap characters are also added
//          the alignments are printed if no more characters are to be
//            added to matched arrays

void print(MatrixEntry * tablePtr, MatrixEntry * table,
	   char * matchedAPtr, char * matchedBPtr,
	   char * seqAPtr, char * seqBPtr,
	   long seqALength)
{
  long nuRowGaps, nuColGaps;	// number of row and column gap characters
  MatrixEntry * prevTableEntry; // table entry of prefix
  MatrixEntryPtrOSLSet prevTableSet; // linked list of prefixes
  Pix linkListPosition;		      // position within the linked list of prefixes
  char * mAP, * mBP, * sAP, * sBP;    // pointers for use during recursion
  
  *--matchedAPtr = *--seqAPtr;
  *--matchedBPtr = *--seqBPtr;

  if (allSequences)  {
    prevTableSet = tablePtr->prevList;
    if (prevTableSet.empty())  { // end of alignment
      if (globalAlignment)  {
	determineGapCharacters(tablePtr, (MatrixEntry *) 0, table,
			       seqALength, nuRowGaps, nuColGaps);
	printGapCharacters(nuRowGaps, nuColGaps,
			   matchedAPtr, matchedBPtr, seqAPtr,
			   seqBPtr);
      }
      cout << matchedAPtr << '\n';
      cout << matchedBPtr << '\n';
      cout << '\n';
    }
    else  { // many alignments
      for (linkListPosition = prevTableSet.first();
	   linkListPosition /* != 0 */;
	   prevTableSet.next(linkListPosition))  {
	prevTableEntry = prevTableSet(linkListPosition);
	mAP = matchedAPtr;
	mBP = matchedBPtr;
	sAP = seqAPtr;
	sBP = seqBPtr;
	determineGapCharacters(tablePtr, prevTableEntry, table,
			       seqALength, nuRowGaps, nuColGaps);
	printGapCharacters(nuRowGaps, nuColGaps,
			   mAP, mBP, sAP, sBP);
	print(prevTableEntry, table, mAP, mBP, sAP, sBP, seqALength);
      }
    }
  }
  else  {			// only one alignment
    prevTableEntry = tablePtr->prev;
    if (globalAlignment || (prevTableEntry != (MatrixEntry *) 0))  {
      // Local alignments end with (MatrixEntry *) 0.
      determineGapCharacters(tablePtr, prevTableEntry, table,
			     seqALength, nuRowGaps, nuColGaps);
      printGapCharacters(nuRowGaps, nuColGaps,
			 matchedAPtr, matchedBPtr, seqAPtr, seqBPtr);
    }
    if (prevTableEntry == (MatrixEntry *) 0)  {
      // Print the strings.
      cout << matchedAPtr << '\n';
      cout << matchedBPtr << '\n';
    }
    else			// not left column or top row
      // Recur.
      print(prevTableEntry, table, matchedAPtr, matchedBPtr, seqAPtr,
	    seqBPtr, seqALength);
  }
  return;
}


int main(int argc, char *argv[])
{
#ifdef TIME
  // Record the starting time.
  clock_t start = clock();
#endif // TIME


  ifstream inputFile;		// problem specification
  
  // Process the command line arguments.
  GetOpt getopt (argc, argv, "ae");
  int option_char;
  
  while ((option_char = getopt ()) != EOF)
    switch (option_char)  {
    case 'a': allSequences = !0; break;
    case 'e': endGaps = !0; break;
    case '?': (void) fprintf(stderr, "%s: [-a] [-e] [file]\n", argv[0]);
      exit(1);
      break;
    }
  
  bool inputFileFlag = argc - getopt.optind;
  switch (inputFileFlag)  {
  case 0:
    break;			// Read the standard input.
  case 1:			// Open a file to read.
    inputFile.open(argv[getopt.optind]);
    if (!inputFile)  {		// failed to open the file
      cerr << argv[0] << ": failed to open the file " << argv[getopt.optind]
	<< "\n";
      exit(1);
    }
    break;
  default:			// too many arguments
    cerr << argv[0] <<": [-a] [file]\n";
    exit(1);
    break;
  }
  
  // Initialize the Data Structures.
  istream & IN = inputFileFlag ? (istream &) inputFile : (istream &) cin;
  SeqStruct seqA(IN);
  SeqStruct seqB(IN);
  
  IN >> globalAlignment;
  globalAlignment = !globalAlignment;
  if (endGaps && !globalAlignment)
    endGaps = 0;		// charging for end gaps makes sense
				// only for global alignments
  
  IN >> seqA.gapOpen;
  IN >> seqA.gapExtend;
  IN >> seqB.gapOpen;
  IN >> seqB.gapExtend;
  
  // Read the alphabets, converting the sequences to integral
  // representations.
  long alphabetALength, alphabetBLength; // number of characters for
  // alphabets of sequences A
  // and B
  int alphaAToIntMap[1<<8]; // map from seq. A alphabet
  // characters to integers
  int alphaBToIntMap[1<<8]; // map from seq. B alphabet
  // characters to integers
  char alphaChar; // alphabet character
  int tmpIndex;	  // index
  int nuCharsUnread; // number of alphabet
  // characters not yet read
  char *charPtr; // character being converted
  int *intPtr;	 // pointer to where value is placed
  
  for (tmpIndex = (1<<8)-1; tmpIndex >= 0; tmpIndex--)
    alphaAToIntMap[tmpIndex] = alphaBToIntMap[tmpIndex] = 0;
  IN >> alphabetALength >> ws;
  for (nuCharsUnread = 0;
       nuCharsUnread < alphabetALength && (alphaChar = IN.get());
       nuCharsUnread++)
    if (!alphaChar)  {
      cerr << argv[0] << ": unexpected end of file\n";
      exit(1);
    }
    else
      alphaAToIntMap[alphaChar] = nuCharsUnread;
  IN >> ws;
  for (tmpIndex = seqA.seqLength-1, charPtr = &(seqA.seq[tmpIndex]),
       intPtr = &(seqA.intSeq[tmpIndex]);
       tmpIndex >= 0;
       tmpIndex--, charPtr--, intPtr--)
    *intPtr = alphaAToIntMap[*charPtr];
  
  IN >> alphabetBLength >> ws;
  for (nuCharsUnread = 0;
       nuCharsUnread < alphabetBLength && (alphaChar = IN.get());
       nuCharsUnread++)
    if (!alphaChar)  {
      cerr << argv[0] << ": unexpected end of file\n";
      exit(1);
    }
    else
      alphaBToIntMap[alphaChar] = nuCharsUnread;
  IN >> ws;
  for (tmpIndex = seqB.seqLength-1, charPtr = &(seqB.seq[tmpIndex]),
       intPtr = &(seqB.intSeq[tmpIndex]);
       tmpIndex >= 0;
       tmpIndex--, charPtr--, intPtr--)
    *intPtr = alphaBToIntMap[*charPtr];
  alphabetBBits = nuBits(alphabetBLength);
  
  // Construct the "match" array.
  double *match = new double[alphabetALength * (1 << alphabetBBits)];
  int numberOfRowsToRead;
  long rowNumber, colNumber;
  char ltrA, ltrB;
  double value;
  for (numberOfRowsToRead = alphabetALength * alphabetBLength;
       numberOfRowsToRead > 0;
       numberOfRowsToRead--)  {
    IN >> rowNumber >> colNumber >> ltrA >> ltrB >> value;
    match[matchLocation(rowNumber-1, colNumber-1)] = value;
  }
  
  // Close the input file.
  if (inputFileFlag) // close the input file
    inputFile.close();


  // Complete the "table."
  // EXCEPTION: End gap penalties are not applied to the bottom row
  // and right column until calculating the global alignment's maximum.
  long seqALength = seqA.seqLength;
  long seqBLength = seqB.seqLength;
  double aGapPenalty;		// gap penalty for sequence A
  double bGapPenalty;		// gap penalty for sequence B
  double seqAGapOpen = seqA.gapOpen;
  double seqBGapOpen = seqB.gapOpen;
  double seqAGapExtend = seqA.gapExtend;
  double seqBGapExtend = seqB.gapExtend;
  
  MatrixEntry *table = new MatrixEntry[seqBLength * seqALength];
  MatrixEntry *prevCol = new MatrixEntry[seqALength];
  MatrixEntry prevRow;
  MatrixEntry *currentTableEntry; // current position in "table"
  MatrixEntry *diagonalTableEntry; // table[i-1,j-1]
  MatrixEntry *prevColEntry;	   // current position in "prevCol"
  int *currentACharacter;	   // current character in seqA
  int *currentBCharacter;	   // current character in seqB
  double maximumValue;		   // value of the maximum
  double previousValue;		   // previous value of the score
				   // being modified
  
  currentTableEntry = table;
  currentBCharacter = seqB.intSeq;
  
  // Complete table[0,0].
  currentACharacter = seqA.intSeq;
  currentTableEntry->score =
    (globalAlignment ?
     match[matchLocation(*currentACharacter, *currentBCharacter)] :
     max(match[matchLocation(*currentACharacter, *currentBCharacter)], 0.0));
  if (allSequences)
    currentTableEntry->prevList.clear();
  else
    currentTableEntry->prev = (MatrixEntry *) 0;
  
  // Complete table[0,j], j > 0.
  for (currentACharacter++, aGapPenalty = seqAGapOpen,
       currentTableEntry++, prevColEntry = prevCol;
       
       *currentACharacter >= 0;
       
       currentACharacter++, aGapPenalty += seqAGapExtend,
       currentTableEntry++, prevColEntry++)  {
    if (!globalAlignment)
      currentTableEntry->score =
	max(match[matchLocation(*currentACharacter, *currentBCharacter)], 0.0);
    else
      currentTableEntry->score =
	(endGaps ?
	 match[matchLocation(*currentACharacter, *currentBCharacter)]
	 - aGapPenalty :
	 match[matchLocation(*currentACharacter, *currentBCharacter)]);
    if (allSequences)
      currentTableEntry->prevList.clear();
    else
      currentTableEntry->prev = (MatrixEntry *) 0;
    prevColEntry->score = (globalAlignment ? -DBL_MAX : 0.0);
    if (allSequences)
      prevColEntry->prevList.clear();
    else
      prevColEntry->prev = (MatrixEntry *) 0;
  }    
  
  // Complete table[i,j], i > 0, no restriction on j.
  bGapPenalty = seqBGapOpen;
  diagonalTableEntry = table - 1;
  for (currentBCharacter++;
       *currentBCharacter >= 0;
       currentBCharacter++, bGapPenalty += seqBGapExtend)  {
    
    // Complete table[i,0], i > 0.
    currentACharacter = seqA.intSeq;
    if (!globalAlignment)
      currentTableEntry->score =
	max(match[matchLocation(*currentACharacter, *currentBCharacter)], 0.0);
    else
      currentTableEntry->score =
      (endGaps ? 
       match[matchLocation(*currentACharacter, *currentBCharacter)] -
       bGapPenalty :
       match[matchLocation(*currentACharacter, *currentBCharacter)]);
    if (allSequences)
      currentTableEntry->prevList.clear();
    else
      currentTableEntry->prev = (MatrixEntry *) 0;
    prevRow.score = (globalAlignment ? -DBL_MAX : 0.0);
    if (allSequences)
      prevRow.prevList.clear();
    else
      prevRow.prev = (MatrixEntry *) 0;
    
    // Complete table[i,j], fixed i > 0, j=1..end.
    for (currentACharacter++, currentTableEntry++, diagonalTableEntry++,
	 prevColEntry = prevCol;
	 
	 *currentACharacter >= 0;
	 
	 currentACharacter++, currentTableEntry++,
	 diagonalTableEntry++, prevColEntry++)  {
      // Compute the table's entry.
      maximumValue = max3(diagonalTableEntry->score,
			  prevColEntry->score, prevRow.score);
      currentTableEntry->score = 
	(globalAlignment ?
	 match[matchLocation(*currentACharacter, *currentBCharacter)]
	 + maximumValue :
	 max(match[matchLocation(*currentACharacter, *currentBCharacter)]
	     + maximumValue, 0.0));
      if (allSequences)  {
	currentTableEntry->prevList.clear();
	if (globalAlignment || maximumValue != 0.0)  {
	  if (maximumValue == diagonalTableEntry->score)
	    (void) currentTableEntry->prevList.add(diagonalTableEntry);
	  if (maximumValue == prevColEntry->score)
	    copyAndAppend(currentTableEntry->prevList,
			  prevColEntry->prevList);
	  if (maximumValue == prevRow.score)
	    copyAndAppend(currentTableEntry->prevList,
			  prevRow.prevList);
	}
      }
      else  {			// only one entry
	if (maximumValue == 0.0 && !globalAlignment)
	  currentTableEntry->prev = (MatrixEntry *) 0;
	else if (maximumValue == diagonalTableEntry->score)
	  currentTableEntry->prev = diagonalTableEntry;
	else if (maximumValue == prevColEntry->score)
	  currentTableEntry->prev = prevColEntry->prev;
	else
	  currentTableEntry->prev = prevRow.prev;
      }
      
      // Compute the prevRow entry.
      previousValue = prevRow.score;
      maximumValue = max(diagonalTableEntry->score - seqAGapOpen,
			 previousValue);
      prevRow.score = 
	(globalAlignment ? maximumValue - seqAGapExtend : 
	 max(maximumValue - seqAGapExtend,0.0));
      if (allSequences)  {
	if (!globalAlignment && prevRow.score == 0.0)
	  prevRow.prevList.clear();
	else {
	  if (maximumValue != previousValue)
	    prevRow.prevList.clear();
	  if (fabs(maximumValue - (diagonalTableEntry->score -
				  seqAGapOpen)) <= 3 * DBL_EPSILON)
	    (void) prevRow.prevList.add(diagonalTableEntry);
	}
      }
      else  {			// only one entry
	if (!globalAlignment && prevRow.score == 0.0)
	  prevRow.prevList.clear();
	else if (maximumValue != previousValue)
	    prevRow.prev = diagonalTableEntry;
	// else maximumValue == previousValue
      }
      
      // Compute the prevCol entry.
      previousValue = prevColEntry->score;
      maximumValue = max(diagonalTableEntry->score - seqBGapOpen,
			 previousValue);
      prevColEntry->score = 
	(globalAlignment ? maximumValue - seqBGapExtend :
	 max(maximumValue - seqBGapExtend,0));
      if (allSequences)  {
	if (!globalAlignment && prevColEntry->score == 0.0)
	  prevColEntry->prevList.clear();
	else  {
	  if (maximumValue != previousValue)
	    prevColEntry->prevList.clear();
	  if (fabs(maximumValue - (diagonalTableEntry->score -
				  seqBGapOpen)) <= 3 * DBL_EPSILON)
	    (void) prevColEntry->prevList.add(diagonalTableEntry);
	}
      }
      else  {			// only one entry
	if (!globalAlignment && prevColEntry->score == 0.0)
	  prevColEntry->prevList.clear();
	else if (maximumValue != previousValue)
	    prevColEntry->prev = diagonalTableEntry;
	// else maximumValue == previousValue
      }
    }
  }
  
  
  // Print the matched sequences.
  // Find the maximum score and apply end gap penalties for the bottom
  // row and right column if applicable.
  double maxScore = -DBL_MAX;
  MatrixEntry * sentinelTableEntry; // marks end of search
  if (globalAlignment)  {
    // Search the last column.
    for (currentTableEntry = table + seqALength - 1,
	 aGapPenalty = seqAGapOpen + seqAGapExtend * (seqBLength - 1),
	 sentinelTableEntry = table + seqALength * seqBLength - 1;
	 currentTableEntry < sentinelTableEntry;
	 currentTableEntry += seqALength, aGapPenalty -= seqAGapExtend)  {
      if (endGaps)
	currentTableEntry->score -= aGapPenalty;
      maxScore = max(maxScore, currentTableEntry->score);
    }
    maxScore = max(maxScore, sentinelTableEntry->score);

    // Search the bottom row.
    for (currentTableEntry = table + (seqBLength - 1) * seqALength,
	 bGapPenalty = seqBGapOpen + seqBGapExtend * (seqALength - 1);
	 currentTableEntry < sentinelTableEntry;
	 currentTableEntry++, bGapPenalty -= seqBGapExtend)  {
      if (endGaps)
	currentTableEntry->score -= bGapPenalty;
      maxScore = max(maxScore, currentTableEntry->score);
    }
  }
  else				// local alignment
    for (currentTableEntry = table + seqBLength * seqALength - 1;
	 currentTableEntry >= table;
	 currentTableEntry--)
      maxScore = max(maxScore, currentTableEntry->score);
  cout << maxScore << '\n';
  
  // Print the alignments.
  char * matchA = new char [seqALength + seqBLength + 1];
  char * matchB = new char [seqALength + seqBLength + 1];
  char * seqAPtr;
  char * seqBPtr;
  char * matchAPtr = matchA + seqALength + seqBLength;
  char * matchBPtr = matchB + seqALength + seqBLength;
  *matchAPtr = '\0';
  *matchBPtr = '\0';
  long rowGapNu, colGapNu;	// number of row and column gap characters
  
  bool printed = 0;		// ! =0 if an alignment has been printed
  // row
  if (globalAlignment)  {
    // Search the bottom row.
    for (currentTableEntry = table + seqBLength * seqALength - 1,
	 sentinelTableEntry = table + (seqBLength - 1) * seqALength -1;
				 // marks end of search
	 (allSequences || !printed) && currentTableEntry != sentinelTableEntry;
	 // print only one alignment if desired
	 currentTableEntry--)
      if (fabs(currentTableEntry->score - maxScore) <= 3 * DBL_EPSILON)  {
	printed = !0;
	seqAPtr = seqA.seq + seqALength;
	seqBPtr = seqB.seq + seqBLength;
	matchAPtr = matchA + seqALength + seqBLength;
	matchBPtr = matchB + seqALength + seqBLength;
	rowGapNu = nuRows(table + seqBLength * seqALength - 1,
			  currentTableEntry, seqALength);
	colGapNu = nuColumns(table + seqBLength * seqALength - 1,
			     currentTableEntry, seqALength);
	printGapCharacters(rowGapNu, colGapNu,
			   matchAPtr, matchBPtr, seqAPtr, seqBPtr);
	print(currentTableEntry, table, matchAPtr, matchBPtr, seqAPtr,
	      seqBPtr, seqALength);
      }
    // Search the right column.
    for (currentTableEntry = sentinelTableEntry;
	 (allSequences || !printed) && currentTableEntry >= table;
	 // print only one alignment if desired
	 currentTableEntry -= seqALength)
      if (fabs(currentTableEntry->score - maxScore) <= 3 * DBL_EPSILON)  {
	printed = !0;
	seqAPtr = seqA.seq + seqALength;
	seqBPtr = seqB.seq + seqBLength;
	matchAPtr = matchA + seqALength + seqBLength;
	matchBPtr = matchB + seqALength + seqBLength;
	rowGapNu = nuRows(table + seqBLength * seqALength - 1,
			  currentTableEntry, seqALength);
	colGapNu = nuColumns(table + seqBLength * seqALength - 1,
			     currentTableEntry, seqALength);
	printGapCharacters(rowGapNu, colGapNu,
			   matchAPtr, matchBPtr, seqAPtr, seqBPtr);
	print(currentTableEntry, table, matchAPtr, matchBPtr, seqAPtr,
	      seqBPtr, seqALength);
      }
  }
  else  {			// local alignment
    for (currentTableEntry = table + seqBLength * seqALength - 1;
	 (allSequences || !printed) && currentTableEntry >= table;
	 // print only one alignment if desired
	 currentTableEntry--)
      if (currentTableEntry->score == maxScore)  {
	printed = !0;
	seqAPtr = seqA.seq + nuColumns(currentTableEntry, table, seqALength) + 1;
	seqBPtr = seqB.seq + nuRows(currentTableEntry, table, seqALength) + 1;
	matchAPtr = matchA + seqALength + seqBLength;
	matchBPtr = matchB + seqALength + seqBLength;
	print(currentTableEntry, table, matchAPtr, matchBPtr, seqAPtr,
	      seqBPtr, seqALength);
      }
  }
  
  // Deallocate the memory.
  delete match;
  delete table;
  delete prevCol;
  delete matchA;
  delete matchB;

#ifdef TIME
  // Record the ending time and print the running time,
  clock_t end = clock();
  cerr << "The running time was " <<
    ((double) (end - start)) / CLOCKS_PER_SEC << " seconds.\n";
#endif // TIME
  
  return 0;
}
